/*
    FEARLUS/SPOM 1-1-4: FearlusArguments.m
    Copyright (C) 2007  Macaulay Institute

    This file is part of FEARLUS/SPOM 1-1-4, an agent-based model of land use
    change and stochastic patch occupancy model.

    FEARLUS/SPOM 1-1-4 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS/SPOM 1-1-4 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/* Implementation of the FearlusArguments class -- a class to parse
 * any command line arguments to be implemented by the model. Pretty
 * much all copied from the Arguments page in the Swarm documentation
 * (Example objectbase/Arguments/1)
 *
 * model0-5-3-T: Changed to be a single-instance class to allow other
 * classes to access FEARLUS-specific information about the arguments
 * through the class rather than through the Swarm-provided global
 * variable 'arguments'. This is required to stop the compiler
 * complaining that the arguments variable, which follows the
 * Arguments protocol, doesn't implement (necessarily) any of the
 * FEARLUS-specific methods for argument information.
 */

#import "FearlusArguments.h"
#import "Verbosity.h"
#import "ClassInfo.h"
#import <random.h>
#import <objc/objc-api.h>
#import <string.h>
#import <ctype.h>
#import <sys/types.h>
#import <unistd.h>
//#import <stdlib.h>
#import <misc.h>
#import <time.h>

static FearlusArguments *squabbles;
static BOOL instantiated = NO;

#ifdef SPOM

static const char *parameterFile;

#else

#define DEBUG_PARCELS 0x1
#define DEBUG_USES 0x2
#define DEBUG_MANAGERS 0x4
#define DEBUG_ALLOCATOR 0x8
#define DEBUG_ENVIRONMENT 0x10
#define DEBUG_OBSERVER 0x20
#define DEBUG_MODEL 0x40
#define DEBUG_CORE_METHODS 0x80
#define DEBUG_INNER_CORE_METHODS 0x100	// Neighbourhood & sorting
#define DEBUG_NEIGHBOURHOOD 0x200	// Specific neighbourhood messages
#define DEBUG_GOVERNMENT 0x800

static const char *seedArg;
static BOOL seedSpecified;
static const char *postInitSeedArg;
static BOOL postInitSeedSpecified;
static const char *modelFile;
static BOOL modelFileSpecified;
static const char *reporterFile;
static BOOL reporterFileSpecified;
static const char *reporterCFFile;
static BOOL reporterCFFileSpecified;
static BOOL appendReport;
static const char *observerFile;
static BOOL observerFileSpecified;
static const char *ontologyFile;
static BOOL ontologyFileSpecified;
static const char *ontologyClass;
static BOOL ontologyClassSpecified;
static const char *ontologyURI;
static BOOL ontologyURISpecified;

static BOOL seed1set = NO, seed2set = NO;
static unsigned seed1, seed2;

static BOOL modelOntologyAllYears = NO;

#endif

static BOOL gridServiceURLSpecified;
static const char *gridServiceURL;
static BOOL gridServiceUIDSpecified;
static const char *gridServiceUID;
static BOOL gridModelDescriptionSpecified;
static const char *gridModelDescription;
static BOOL javapathSpecified;
static const char *javapath;

static BOOL rngClassSpecified;
static Class rngClass;

@implementation FearlusArguments

/* +createBegin:
 *
 * This method is copied straight out of the Arguments class page in
 * the Swarm documentation, with minor changes as appropriate.
 */

+createBegin: (id)aZone {
  static struct argp_option options[] = {
#ifdef SPOM
    {"parameter", 'p', "PARAMETER_FILE", 0,
     "Set the name of the parameters file", 99},
#else
    {"withseed", 'X', "SEED", 0, "Specify the seed. [0-9]+, TIME or DEFAULT",
     99},
    {"postinitseed", 'Z', "POSTINITSEED", 0, "Specify a separate seed to use "
     "after initialisation: [0-9]+ or TIME", 99},
    {"parameters", 'p', "PARAMETERFILE", 0,
     "File for the model parameters to be loaded from", 99},
    {"observers", 'o', "OBSERVERFILE", 0,
     "File for the observer settings to be loaded from", 99},
    {"report", 'r', "REPORTOUTPUTFILE", 0,
     "File to save the report to (stdout by default)", 99},
    {"repconfig", 'R', "REPORTCONFIGFILE", 0,
     "Reporter configuration file", 99},
    {"append-report", 'a', NULL, 0, "If report file exists, then append to it",
     99},
    {"ontology", 'O', "ONTOLOGYFILE", 0,
     "Name of file to output ontology to", 99},
    {"ontology-class", 'C', "CLASS", 0,
     "Record structural ontology from subclasses of CLASS", 99},
    {"ontology-uri", 'U', "URI", 0,
     "URI for ontology", 99},
    {"ontology-all-years", 'A', NULL, 0,
     "Output a model state ontology each year (warning: will potentially "
     "require a lot of disk space)", 99},
#endif
    {"warranty", 'w', NULL, 0, "Show warranty information", 999},
    {"conditions", 'c', NULL, 0, "Show conditions of redistribution", 999},
    {"debug", 'D', "DEBUGLEVEL", 0,
     "Debug level (integer) and/or a list of +/- separated message symbols",
     99},
    {"gridServiceURL", 'G', "GRIDURL", 0, "URL for FEARLUS-G Service", 99},
    {"gridServiceUID", 'g', "GRIDUSERID", 0, "User ID for FEARLUS-G Service",
     99},
    {"gridModelDescription", 'H', "GRIDDESCRIPTION", 0,
     "Description for this model", 99},
    {"javapath",'j', "JAVAPATH", 0, "Path to java Grid Client", 99},
    {"rng", 'n', "RNGCLASS", 0, "Class of RNG to use", 99},
    { 0 }
  };

  if(instantiated) {
    fprintf(stderr, "WARNING: createBegin of FearlusArguments called again\n");
  }
  squabbles = [super createBegin: aZone];
  instantiated = YES;

  [squabbles addOptions: options];
  
  #ifdef SPOM
  parameterFile = "SPOMEnvironment.spom";
  #else
  seedSpecified = NO;
  postInitSeedSpecified = NO;
  modelFileSpecified = NO;
  modelFile = "fearlus.model";
  reporterFileSpecified = NO;
  reporterFile = "fearlus-report.txt";
  reporterCFFileSpecified = NO;
  reporterCFFile = "fearlus-report.cf";
  appendReport = NO;
  observerFileSpecified = NO;
  observerFile = "fearlus.obs";
  ontologyFileSpecified = NO;
  ontologyFile = "fearlus.owl";
  ontologyClassSpecified = NO;
  ontologyClass = "SwarmObject";
  ontologyURISpecified = NO;
  ontologyURI = NULL;
  #endif
  
  gridServiceURLSpecified = NO;
  gridServiceUIDSpecified = NO;
  gridModelDescriptionSpecified = NO;
  javapathSpecified = NO;

  rngClassSpecified = NO;

  return squabbles;
}

/* -parseKey:arg:
 *
 * Parse the argument for the given key. Switch used to enable easy
 * expansion.
 */

-(int)parseKey: (int)key arg: (const char *)arg {
  switch(key) {
  case 'D':
    [Verbosity setVerbosity: arg];
    return 0;
  case 'G':
    gridServiceURL = arg;
    gridServiceURLSpecified = YES;
    return 0;
  case 'g':
    gridServiceUID = arg;
    gridServiceUIDSpecified = YES;
    return 0;
  case 'H':
    gridModelDescription = arg;
    gridModelDescriptionSpecified = YES;
    return 0;
  case 'j':
    javapath = arg;
    javapathSpecified = YES;
    return 0;
  case 'n':
    rngClass = objc_lookup_class(arg);
    if(rngClass == Nil) {
      fprintf(stderr, "Invalid class name specified for rng command-line"
	      "option (%s)\n", arg);
      abort();
    }

    rngClassSpecified = YES;
    return 0;
    
#ifdef SPOM

  case 'p':
    parameterFile = arg;
    return 0;

#else

  case 'X':
    seedArg = arg;
    seedSpecified = YES;
    return 0;
  case 'Z':
    postInitSeedArg = arg;
    postInitSeedSpecified = YES;
    return 0;
  case 'o':
    observerFile = arg;
    observerFileSpecified = YES;
    return 0;
  case 'p':
    modelFile = arg;
    modelFileSpecified = YES;
    return 0;
  case 'r':
    reporterFile = arg;
    reporterFileSpecified = YES;
    return 0;
  case 'R':
    reporterCFFile = arg;
    reporterCFFileSpecified = YES;
    return 0;
  case 'a':
    appendReport = YES;
    return 0;

  case 'O':
    ontologyFile = arg;
    ontologyFileSpecified = YES;
    return 0;
  case 'C':
    if(objc_lookup_class(arg) == Nil) {
      fprintf(stderr, "Class %s is not a recognised class\n", arg);
      abort();
    }
    if(![ClassInfo class: objc_get_class(arg)
		   isSubClassOf: [SwarmObject class]]) {
      fprintf(stderr, "Class %s is not a kind of SwarmObject\n", arg);
      abort();
    }
    ontologyClass = arg;
    ontologyClassSpecified = YES;
    return 0;
  case 'U':
    ontologyURI = arg;
    ontologyURISpecified = YES;
    return 0;
  case 'A':
    modelOntologyAllYears = YES;
    return 0;
#endif

  case 'c':
    printf(
"		    GNU GENERAL PUBLIC LICENSE\n"
"   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION\n"
"\n"
"  0. This License applies to any program or other work which contains\n"
"a notice placed by the copyright holder saying it may be distributed\n"
"under the terms of this General Public License.  The \"Program\", below,\n"
"refers to any such program or work, and a \"work based on the Program\"\n"
"means either the Program or any derivative work under copyright law:\n"
"that is to say, a work containing the Program or a portion of it,\n"
"either verbatim or with modifications and/or translated into another\n"
"language.  (Hereinafter, translation is included without limitation in\n"
"the term \"modification\".)  Each licensee is addressed as \"you\".\n"
"\n"
"Activities other than copying, distribution and modification are not\n"
"covered by this License; they are outside its scope.  The act of\n"
"running the Program is not restricted, and the output from the Program\n"
"is covered only if its contents constitute a work based on the\n"
"Program (independent of having been made by running the Program).\n"
"Whether that is true depends on what the Program does.\n"
"\n"
"  1. You may copy and distribute verbatim copies of the Program's\n"
"source code as you receive it, in any medium, provided that you\n"
"conspicuously and appropriately publish on each copy an appropriate\n"
"copyright notice and disclaimer of warranty; keep intact all the\n"
"notices that refer to this License and to the absence of any warranty;\n"
"and give any other recipients of the Program a copy of this License\n"
"along with the Program.\n"
"\n"
"You may charge a fee for the physical act of transferring a copy, and\n"
"you may at your option offer warranty protection in exchange for a fee.\n"
"\n"
"  2. You may modify your copy or copies of the Program or any portion\n"
"of it, thus forming a work based on the Program, and copy and\n"
"distribute such modifications or work under the terms of Section 1\n"
"above, provided that you also meet all of these conditions:\n"
"\n"
"    a) You must cause the modified files to carry prominent notices\n"
"    stating that you changed the files and the date of any change.\n"
"\n"
"    b) You must cause any work that you distribute or publish, that in\n"
"    whole or in part contains or is derived from the Program or any\n"
"    part thereof, to be licensed as a whole at no charge to all third\n"
"    parties under the terms of this License.\n"
"\n"
"    c) If the modified program normally reads commands interactively\n"
"    when run, you must cause it, when started running for such\n"
"    interactive use in the most ordinary way, to print or display an\n"
"    announcement including an appropriate copyright notice and a\n"
"    notice that there is no warranty (or else, saying that you provide\n"
"    a warranty) and that users may redistribute the program under\n"
"    these conditions, and telling the user how to view a copy of this\n"
"    License.  (Exception: if the Program itself is interactive but\n"
"    does not normally print such an announcement, your work based on\n"
"    the Program is not required to print an announcement.)\n"
"\n"
"These requirements apply to the modified work as a whole.  If\n"
"identifiable sections of that work are not derived from the Program,\n"
"and can be reasonably considered independent and separate works in\n"
"themselves, then this License, and its terms, do not apply to those\n"
"sections when you distribute them as separate works.  But when you\n"
"distribute the same sections as part of a whole which is a work based\n"
"on the Program, the distribution of the whole must be on the terms of\n"
"this License, whose permissions for other licensees extend to the\n"
"entire whole, and thus to each and every part regardless of who wrote it.\n"
"\n"
"Thus, it is not the intent of this section to claim rights or contest\n"
"your rights to work written entirely by you; rather, the intent is to\n"
"exercise the right to control the distribution of derivative or\n"
"collective works based on the Program.\n"
"\n"
"In addition, mere aggregation of another work not based on the Program\n"
"with the Program (or with a work based on the Program) on a volume of\n"
"a storage or distribution medium does not bring the other work under\n"
"the scope of this License.\n"
"\n"
"  3. You may copy and distribute the Program (or a work based on it,\n"
"under Section 2) in object code or executable form under the terms of\n"
"Sections 1 and 2 above provided that you also do one of the following:\n"
"\n"
"    a) Accompany it with the complete corresponding machine-readable\n"
"    source code, which must be distributed under the terms of Sections\n"
"    1 and 2 above on a medium customarily used for software interchange; or,\n"
"\n"
"    b) Accompany it with a written offer, valid for at least three\n"
"    years, to give any third party, for a charge no more than your\n"
"    cost of physically performing source distribution, a complete\n"
"    machine-readable copy of the corresponding source code, to be\n"
"    distributed under the terms of Sections 1 and 2 above on a medium\n"
"    customarily used for software interchange; or,\n"
"\n"
"    c) Accompany it with the information you received as to the offer\n"
"    to distribute corresponding source code.  (This alternative is\n"
"    allowed only for noncommercial distribution and only if you\n"
"    received the program in object code or executable form with such\n"
"    an offer, in accord with Subsection b above.)\n"
"\n"
"The source code for a work means the preferred form of the work for\n"
"making modifications to it.  For an executable work, complete source\n"
"code means all the source code for all modules it contains, plus any\n"
"associated interface definition files, plus the scripts used to\n"
"control compilation and installation of the executable.  However, as a\n"
"special exception, the source code distributed need not include\n"
"anything that is normally distributed (in either source or binary\n"
"form) with the major components (compiler, kernel, and so on) of the\n"
"operating system on which the executable runs, unless that component\n"
"itself accompanies the executable.\n"
"\n"
"If distribution of executable or object code is made by offering\n"
"access to copy from a designated place, then offering equivalent\n"
"access to copy the source code from the same place counts as\n"
"distribution of the source code, even though third parties are not\n"
"compelled to copy the source along with the object code.\n"
"\n"
"  4. You may not copy, modify, sublicense, or distribute the Program\n"
"except as expressly provided under this License.  Any attempt\n"
"otherwise to copy, modify, sublicense or distribute the Program is\n"
"void, and will automatically terminate your rights under this License.\n"
"However, parties who have received copies, or rights, from you under\n"
"this License will not have their licenses terminated so long as such\n"
"parties remain in full compliance.\n"
"\n"
"  5. You are not required to accept this License, since you have not\n"
"signed it.  However, nothing else grants you permission to modify or\n"
"distribute the Program or its derivative works.  These actions are\n"
"prohibited by law if you do not accept this License.  Therefore, by\n"
"modifying or distributing the Program (or any work based on the\n"
"Program), you indicate your acceptance of this License to do so, and\n"
"all its terms and conditions for copying, distributing or modifying\n"
"the Program or works based on it.\n"
"\n"
"  6. Each time you redistribute the Program (or any work based on the\n"
"Program), the recipient automatically receives a license from the\n"
"original licensor to copy, distribute or modify the Program subject to\n"
"these terms and conditions.  You may not impose any further\n"
"restrictions on the recipients' exercise of the rights granted herein.\n"
"You are not responsible for enforcing compliance by third parties to\n"
"this License.\n"
"\n"
"  7. If, as a consequence of a court judgment or allegation of patent\n"
"infringement or for any other reason (not limited to patent issues),\n"
"conditions are imposed on you (whether by court order, agreement or\n"
"otherwise) that contradict the conditions of this License, they do not\n"
"excuse you from the conditions of this License.  If you cannot\n"
"distribute so as to satisfy simultaneously your obligations under this\n"
"License and any other pertinent obligations, then as a consequence you\n"
"may not distribute the Program at all.  For example, if a patent\n"
"license would not permit royalty-free redistribution of the Program by\n"
"all those who receive copies directly or indirectly through you, then\n"
"the only way you could satisfy both it and this License would be to\n"
"refrain entirely from distribution of the Program.\n"
"\n"
"If any portion of this section is held invalid or unenforceable under\n"
"any particular circumstance, the balance of the section is intended to\n"
"apply and the section as a whole is intended to apply in other\n"
"circumstances.\n"
"\n"
"It is not the purpose of this section to induce you to infringe any\n"
"patents or other property right claims or to contest validity of any\n"
"such claims; this section has the sole purpose of protecting the\n"
"integrity of the free software distribution system, which is\n"
"implemented by public license practices.  Many people have made\n"
"generous contributions to the wide range of software distributed\n"
"through that system in reliance on consistent application of that\n"
"system; it is up to the author/donor to decide if he or she is willing\n"
"to distribute software through any other system and a licensee cannot\n"
"impose that choice.\n"
"\n"
"This section is intended to make thoroughly clear what is believed to\n"
"be a consequence of the rest of this License.\n"
"\n"
"  8. If the distribution and/or use of the Program is restricted in\n"
"certain countries either by patents or by copyrighted interfaces, the\n"
"original copyright holder who places the Program under this License\n"
"may add an explicit geographical distribution limitation excluding\n"
"those countries, so that distribution is permitted only in or among\n"
"countries not thus excluded.  In such case, this License incorporates\n"
"the limitation as if written in the body of this License.\n"
"\n"
"  9. The Free Software Foundation may publish revised and/or new versions\n"
"of the General Public License from time to time.  Such new versions will\n"
"be similar in spirit to the present version, but may differ in detail to\n"
"address new problems or concerns.\n"
"\n"
"Each version is given a distinguishing version number.  If the Program\n"
"specifies a version number of this License which applies to it and \"any\n"
"later version\", you have the option of following the terms and conditions\n"
"either of that version or of any later version published by the Free\n"
"Software Foundation.  If the Program does not specify a version number of\n"
"this License, you may choose any version ever published by the Free Software\n"
"Foundation.\n"
"\n"
"  10. If you wish to incorporate parts of the Program into other free\n"
"programs whose distribution conditions are different, write to the author\n"
"to ask for permission.  For software which is copyrighted by the Free\n"
"Software Foundation, write to the Free Software Foundation; we sometimes\n"
"make exceptions for this.  Our decision will be guided by the two goals\n"
"of preserving the free status of all derivatives of our free software and\n"
"of promoting the sharing and reuse of software generally.\n"
);
    abort();
  case 'w':
    printf("			    NO WARRANTY\n"
	   "\n"
	   "11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO "
	   "WARRANTY\n"
	   "FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  "
	   "EXCEPT WHEN\n"
	   "OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER "
	   "PARTIES\n"
	   "PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, "
	   "EITHER EXPRESSED\n"
	   "OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES "
	   "OF\n"
	   "MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE "
	   "RISK AS\n"
	   "TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  "
	   "SHOULD THE\n"
	   "PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY "
	   "SERVICING,\n"
	   "REPAIR OR CORRECTION.\n"
	   "\n"
	   "12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN "
	   "WRITING\n"
	   "WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY "
	   "AND/OR\n"
	   "REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR "
	   "DAMAGES,\n"
	   "INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL "
	   "DAMAGES ARISING\n"
	   "OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT "
	   "LIMITED\n"
	   "TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES "
	   "SUSTAINED BY\n"
	   "YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH "
	   "ANY OTHER\n"
	   "PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF "
	   "THE\n"
	   "POSSIBILITY OF SUCH DAMAGES.\n");
    abort();

  default:
    return [super parseKey: key arg: arg];
  }
}


#ifdef SPOM

/* +getParameterFile
 *
 * Return the name of the parameter file
 */
 
+(const char *)getParameterFile {
  return parameterFile;
}

#else

/* +seedHasBeenSpecified
 *
 * Return a boolean value specifying whether or not a seed has been
 * specified on the command line argument. Returns true for [0-9]+ or
 * TIME values, false otherwise.
 */

+(BOOL)seedHasBeenSpecified {
  if(seedSpecified) {
    if(!strcmp(seedArg, "TIME")) {
      return YES;
    }
    else if(!strcmp(seedArg, "DEFAULT")) {
      return NO;
    }
    else {
      const char *p = seedArg;

      while(*p) {
	if(!isdigit((int)*p)) {
	  fprintf(stderr, "Invalid value for seed argument: %s -- using "
		  "default value\n", seedArg);
	  return NO;
	}
	p++;
      }
      return YES;
    }
  }
  else {
    return NO;
  }
}

/* +getSeedArg
 *
 * Return the seed argument value -- this is either the number
 * specified, or if the seedArg is "TIME", then use the current
 * internal clock time as a seed.
 */

+(unsigned)getSeedArg {
  unsigned seed;

  if(!seedSpecified || !strcmp(seedArg, "DEFAULT")) {
    fprintf(stderr, "Cannot get the seed value when it isn't specified!\n");
    return 0;
  }
  if(!strcmp(seedArg, "TIME")) {
    time_t now = time(0);

    seed = (unsigned)now;
    if(seed2set && seed == seed2) {
      seed++;
    }
    seed1 = seed;
    seed1set = YES;
  }
  else {
    seed = (unsigned)strtoul(seedArg, NULL, 10);
  }
  return seed;
}

/* +postInitseedHasBeenSpecified
 *
 * Return a boolean value specifying whether or not a separate seed to
 * use after initialisation has been given on the command line. Bomb
 * out if an invalid argument has been given.
 */

+(BOOL)postInitSeedHasBeenSpecified {
  if(postInitSeedSpecified) {
    if(!strcmp(postInitSeedArg, "TIME")) {
      return YES;
    }
    else {
      const char *p = postInitSeedArg;

      while(*p) {
	if(!isdigit((int)*p)) {
	  fprintf(stderr, "Invalid value for seed argument: %s\n",
		  postInitSeedArg);
	  abort();
	}
	p++;
      }
      return YES;
    }
  }
  else {
    return NO;
  }
}

/* +getPostInitSeedArg
 *
 * Return the separate seed argument value -- this is either the
 * number specified, or if the seedArg is "TIME", then a value
 * computed from the result of the current clock time -- note that
 * this needs to be different from that used by getSeedArg, so we
 * multiply by microseconds.
 */

+(unsigned)getPostInitSeedArg {
  unsigned seed;

  if(!postInitSeedSpecified) {
    fprintf(stderr, "Cannot get the seed value when it isn't specified!\n");
    abort();
  }
  if(!strcmp(postInitSeedArg, "TIME")) {
    struct timeval now;

    if(gettimeofday(&now, NULL) < 0) {
      perror("Setting seed: gettimeofday");
      abort();
    }
    
    now.tv_usec = (now.tv_usec == 0 ? now.tv_usec + 1 : now.tv_usec);
    seed = (unsigned)now.tv_sec * (unsigned)now.tv_usec;
    if(seed1set && seed == seed1) {
      seed++;
    }
    seed2 = seed;
    seed2set = YES;
  }
  else {
    seed = (unsigned)strtoul(postInitSeedArg, NULL, 10);
  }
  return seed;
}

/* +modelFileHasBeenSpecified
 *
 * Return whether or not the model file was given on the command line.
 */

+(BOOL)modelFileHasBeenSpecified {
  return modelFileSpecified;
}

/* +getModelFile
 *
 * Return the model file to check for model parameters (fearlus.model
 * by default)
 */

+(const char *)getModelFile {
  return modelFile;
}

/* +observerFileHasBeenSpecified
 *
 * Return a value indicating whether or not the observer file was
 * specified on the command line. This is used to determine whether or
 * not to load observer settings.
 */

+(BOOL)observerFileHasBeenSpecified {
  return observerFileSpecified;
}

/* +getObserverFile
 *
 * Return the name passed in as the observer file to use
 */

+(const char *)getObserverFile {
  return observerFile;
}

/* +reporterFileHasBeenSpecified
 *
 * Return a Boolean value indicating whether or not a reporter file
 * has been specified on the command line.
 */

+(BOOL)reporterFileHasBeenSpecified {
  return reporterFileSpecified;
}

/* +getReporterFile
 *
 * Return the reporter file. If none has been specified on the command
 * line, then return the default.
 */

+(const char *)getReporterFile {
  return reporterFile;
}

/* +reporterCFFileHasBeenSpecified
 *
 * Return a Boolean value indicating whether or not a reporter
 * configuration file has been specified on the command line. 
 */

+(BOOL)reporterCFFileHasBeenSpecified {
  return reporterCFFileSpecified;
}

/* +getReporterCFFile
 *
 * Return the reporter configuration file. If none has been specified
 * on the command line, then return the default.
 */

+(const char *)getReporterCFFile {
  return reporterCFFile;
}

/* +ontologyFileHasBeenSpecified
 *
 * Return whether or not an ontology file was given on the command 
 * line.
 */

+(BOOL)ontologyFileHasBeenSpecified {
  return ontologyFileSpecified;
}

/* +getOntologyFile
 *
 * Return the ontology file given on the command line
 */

+(const char *)getOntologyFile {
  return ontologyFile;
}

/* +ontologyClassHasBeenSpecified
 *
 * Return whether or not a top level class for a structural
 * ontology has been specified on the command line.
 */

+(BOOL)ontologyClassHasBeenSpecified {
  return ontologyClassSpecified;
}

/* +getOntologyClass
 *
 * Return the top level class for the structural ontology
 * specified on the command line
 */

+(const char *)getOntologyClass {
  return ontologyClass;
}

/* +ontologyURIHasBeenSpecified
 *
 * Return whether or not a specific URI to use for the 
 * ontology has been specified on the command line
 */

+(BOOL)ontologyURIHasBeenSpecified {
  return ontologyURISpecified;
}

/* +getOntologyURI
 *
 * Return the ontology URI that was specified on the command line
 * Note that it is an error to call this method if none was specified
 */

+(const char *)getOntologyURI {
  if(!ontologyURISpecified) {
    fprintf(stderr, "PANIC! file: %s line: %d\n", __FILE__, __LINE__);
				// Not really a panic, but a bug...
				// We've asked for the ontologyURI
				// when none was given on the command line
    abort();
  }

  return ontologyURI;
}

/* +useReporter
 *
 * Return a Boolean value indicating whether or not a reporter has
 * been requested.  This will be the case if either the reporter
 * output or configuration files or both have been specified. (The
 * default output is stdout.)
 */

+(BOOL)useReporter {
  if(reporterCFFileSpecified || reporterFileSpecified) {
    return YES;
  }
  else {
    return NO;
  }
}

/* +appendReportFile
 *
 * Return a Boolean value indicating whether or not the report file is
 * to be appended to if specified.
 */

+(BOOL)appendReportFile {
  return appendReport;
}

/* +modelStateAllYears
 *
 * Whether or not to output a model state ontology each year in batch mode
 */

+(BOOL)modelStateAllYears {
  return modelOntologyAllYears;
}

#endif

/* +gridServiceURLHasBeenSpecified
 *
 * Whether or not a grid service URL has been specified
 */

+(BOOL)gridServiceURLHasBeenSpecified {
  return gridServiceURLSpecified;
}

/* +getGridServiceURL
 *
 * Return the grid service URL
 */

+(const char *)getGridServiceURL {
  if(!gridServiceURLSpecified) {
    fprintf(stderr, "No grid service URL specified on the command line\n");
    abort();
  }
  return gridServiceURL;
}

/* +gridServiceUIDHasBeenSpecified
 *
 * Whether or not a UID for the grid service has been specified
 */

+(BOOL)gridServiceUIDHasBeenSpecified {
  return gridServiceUIDSpecified;
}

/* +getGridServiceUID
 *
 * The grid service UID specified
 */

+(const char *)getGridServiceUID {
  if(!gridServiceUIDSpecified) {
    fprintf(stderr, "No grid service UID specified on the command line\n");
    abort();
  }
  return gridServiceUID;
}

/* +gridModelDescriptionHasBeenSpecified
 *
 * Whether a grid model description has been specified
 */

+(BOOL)gridModelDescriptionHasBeenSpecified {
  return gridModelDescriptionSpecified;
}

/* +getGridModelDescription
 *
 * The grid model description specified
 */

+(const char *)getGridModelDescription {
  if(!gridModelDescriptionSpecified) {
    fprintf(stderr, "No grid service model description specified on the "
	    "command line\n");
    abort();
  }
  return gridModelDescription;
}

/* +javapathHasBeenSpecified
 *
 * Whether a java path has been specified for the grid service
 */

+(BOOL)javapathHasBeenSpecified {
  return javapathSpecified;
}

/* +getJavapath
 *
 * The java path specified
 */

+(const char *)getJavapath {
  if(!javapathSpecified) {
    fprintf(stderr, "No grid service java path specified on the command "
	    "line\n");
    abort();
  }
  return javapath;
}

/* +rngClassHasBeenSpecified
 *
 * Return whether a class of RNG to use has been specified
 */

+(BOOL)rngClassHasBeenSpecified {
  return rngClassSpecified;
}

/* +getRngClass
 *
 * Return the RNG class the user specified, checking it exists and is valid
 */

+(Class)getRngClass {
  if(![rngClass conformsTo: @protocol(SimpleRandomGenerator)]) {
    fprintf(stderr, "Invalid random number generator class %s\n",
	    [rngClass name]);
    abort();
  }
  return rngClass;
}

@end

